'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.throwTypeErrorIfKeysInvalid = exports.getSameCrioIfUnchanged = exports.getPlainObject = exports.getDeeplyNestedValue = exports.getCleanValue = exports.freezeIfNotProduction = exports.createPrototypeObject = undefined;

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _hashIt = require('hash-it');

var _hashIt2 = _interopRequireDefault(_hashIt);

var _isArray = require('lodash/isArray');

var _isArray2 = _interopRequireDefault(_isArray);

var _constants = require('./constants');

var _is = require('./is');

var _loops = require('./loops');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } } // external dependencies


// utils


/**
 * build prototype object to add to default prototype
 *
 * @param {Object} prototype
 * @returns {Object}
 */
var createPrototypeObject = function createPrototypeObject(prototype) {
  var keys = (0, _constants.OBJECT_KEYS)(prototype);
  var propertySymbols = _constants.OBJECT.getOwnPropertySymbols(prototype);
  var allPropertyItems = [].concat(_toConsumableArray(keys), _toConsumableArray(propertySymbols));

  return allPropertyItems.reduce(function (accumulatedPrototype, key) {
    var value = prototype[key];

    return _extends({}, accumulatedPrototype, _defineProperty({}, key, {
      configurable: true,
      enumerable: false,
      value: value,
      writable: true
    }));
  }, {});
};

/**
 * run Object.freeze on the crio only in non-production environments
 *
 * @param {CrioArray|CrioObject} crio
 * @returns {CrioArray|CrioObject}
 */
var freezeIfNotProduction = function freezeIfNotProduction(crio) {
  if (_constants.IS_PRODUCTION) {
    return crio;
  }

  return Object.freeze(crio);
};

/**
 * get the plain object version of the value passed
 *
 * @param {Array<*>|Object|CrioArray|CrioObject} value
 * @returns {Array<*>|Object}
 */
var getCleanValue = function getCleanValue(value) {
  return (0, _is.isCrio)(value) ? value.thaw() : value;
};

/**
 * get the value for setIn
 *
 * @param {*} currentValue
 * @param {*} value
 * @param {boolean} isMatchingKey
 * @param {Array<string>} restOfKeys
 * @returns {*}
 */
var getDeeplyNestedValue = function getDeeplyNestedValue(currentValue, value, isMatchingKey, restOfKeys) {
  if (!isMatchingKey) {
    return currentValue;
  }

  return (0, _is.isCrio)(currentValue) ? currentValue.setIn(restOfKeys, value) : (0, _loops.createDeeplyNestedObject)(restOfKeys, value);
};

/**
 * get the plain object version of the crio type
 *
 * @param {CrioArray|CrioObject} crio
 * @param {number} crio.length
 * @param {boolean} isDynamicLength=true
 * @returns {Array|Object}
 */
var getPlainObject = function getPlainObject(crio) {
  var isDynamicLength = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;

  if ((0, _is.isCrioArray)(crio)) {
    return isDynamicLength ? [] : new Array(crio.length);
  }

  return {};
};

/**
 * return the original object if the values have not changed
 *
 * @param {CrioArray|CrioObject} crio
 * @param {Array<*>|Object} potentialCrio
 * @returns {CrioArray|CrioObject}
 */
var getSameCrioIfUnchanged = function getSameCrioIfUnchanged(crio, potentialCrio) {
  var hashCode = (0, _hashIt2.default)(potentialCrio);

  if (crio[_constants.CRIO_HASH_CODE] === hashCode) {
    return crio;
  }

  return new crio.constructor(potentialCrio, hashCode);
};

/**
 * if keys passed are not valid, throw a TypeError
 *
 * @param {Array<number|string>} keys
 */
var throwTypeErrorIfKeysInvalid = function throwTypeErrorIfKeysInvalid(keys) {
  if (!(0, _isArray2.default)(keys)) {
    throw new TypeError('Keys passed must be an array.');
  }
};

exports.createPrototypeObject = createPrototypeObject;
exports.freezeIfNotProduction = freezeIfNotProduction;
exports.getCleanValue = getCleanValue;
exports.getDeeplyNestedValue = getDeeplyNestedValue;
exports.getPlainObject = getPlainObject;
exports.getSameCrioIfUnchanged = getSameCrioIfUnchanged;
exports.throwTypeErrorIfKeysInvalid = throwTypeErrorIfKeysInvalid;